@@ -32,16 +32,13 @@
 #include <grp.h>
 #include <pwd.h>
 
-#ifdef  HAVE_LOGINDEVPERM
-#include <libdevinfo.h>
-#endif  /* HAVE_LOGINDEVPERM */
-
 #include <security/pam_appl.h>
 
 #include <glib.h>
 #include <glib/gi18n.h>
 #include <glib/gstdio.h>
 #include <glib-object.h>
+#include <gio/gio.h>
 #define DBUS_API_SUBJECT_TO_CHANGE
 #include <dbus/dbus.h>
 #include <dbus/dbus-glib.h>
@@ -51,6 +48,7 @@
 
 #include "ck-connector.h"
 
+#include "gdm-common.h"
 #include "gdm-session-worker.h"
 #include "gdm-marshal.h"
 
@@ -86,7 +84,7 @@
 #define GDM_SESSION_LOG_FILENAME ".xsession-errors"
 #endif
 
-#define MESSAGE_REPLY_TIMEOUT (10 * 60 * 1000)
+#define MAX_FILE_SIZE     65536
 
 enum {
         GDM_SESSION_WORKER_STATE_NONE = 0,
@@ -94,6 +92,7 @@ enum {
         GDM_SESSION_WORKER_STATE_AUTHENTICATED,
         GDM_SESSION_WORKER_STATE_AUTHORIZED,
         GDM_SESSION_WORKER_STATE_ACCREDITED,
+        GDM_SESSION_WORKER_STATE_ACCOUNT_DETAILS_FETCHED,
         GDM_SESSION_WORKER_STATE_SESSION_OPENED,
         GDM_SESSION_WORKER_STATE_SESSION_STARTED,
         GDM_SESSION_WORKER_STATE_REAUTHENTICATED,
@@ -121,12 +120,12 @@ struct GdmSessionWorkerPrivate
         char             *hostname;
         char             *username;
         uid_t             uid;
+        gid_t             gid;
         gboolean          password_is_required;
 
         int               cred_flags;
 
         char            **arguments;
-        GHashTable       *environment;
         guint32           cancelled : 1;
         guint32           timed_out : 1;
         guint             state_change_idle_id;
@@ -179,6 +178,8 @@ open_ck_session (GdmSessionWorker  *worker)
         const char     *display_hostname;
         gboolean        is_local;
 
+        ret = FALSE;
+
         if (worker->priv->x11_display_name != NULL) {
                 display_name = worker->priv->x11_display_name;
         } else {
@@ -206,9 +207,9 @@ open_ck_session (GdmSessionWorker  *worker)
                 is_local = FALSE;
         }
 
-        pwent = getpwnam (worker->priv->username);
+        gdm_get_pwent_for_name (worker->priv->username, &pwent);
         if (pwent == NULL) {
-                return FALSE;
+                goto out;
         }
 
         worker->priv->ckc = ck_connector_new ();
@@ -590,39 +591,8 @@ static void
 attempt_to_load_user_settings (GdmSessionWorker *worker,
                                const char       *username)
 {
-        struct passwd *passwd_entry;
-        uid_t          old_uid;
-        gid_t          old_gid;
-
-        old_uid = geteuid ();
-        old_gid = getegid ();
-
-        passwd_entry = getpwnam (username);
-
-        /* User input isn't a valid username
-         */
-        if (passwd_entry == NULL) {
-                return;
-        }
-
-        /* We may get called late in the pam conversation after
-         * the user has already been authenticated.  This could
-         * happen if for instance, the user's home directory isn't
-         * available until late in the pam conversation so user
-         * settings couldn't get loaded until late in the conversation.
-         * If we get called late the seteuid/setgid calls here will fail,
-         * but that's okay, because we'll already be the uid/gid we want
-         * to be.
-         */
-        setegid (passwd_entry->pw_gid);
-        seteuid (passwd_entry->pw_uid);
-
         gdm_session_settings_load (worker->priv->user_settings,
-                                   passwd_entry->pw_dir,
-                                   NULL);
-
-        seteuid (old_uid);
-        setegid (old_gid);
+                                   username);
 }
 
 static void
@@ -702,9 +672,14 @@ send_question_method (GdmSessionWorker *worker,
                                         &question);
 
         dbus_error_init (&error);
+
+        /*
+         * Pass in INT_MAX for the timeout.  This is a special value that
+         * means block forever.  This fixes bug #607861
+         */
         reply = dbus_connection_send_with_reply_and_block (worker->priv->connection,
                                                            message,
-                                                           MESSAGE_REPLY_TIMEOUT,
+                                                           INT_MAX,
                                                            &error);
         dbus_message_unref (message);
 
@@ -981,24 +956,8 @@ gdm_session_worker_uninitialize_pam (GdmSessionWorker *worker,
         if (worker->priv->state >= GDM_SESSION_WORKER_STATE_SESSION_OPENED) {
                 pam_close_session (worker->priv->pam_handle, 0);
                 gdm_session_auditor_report_logout (worker->priv->auditor);
-
-#ifdef  HAVE_LOGINDEVPERM
-                /*
-                 * Only do logindevperm processing if /dev/console or
-                 * a device associated with a VT
-                 */
-                if (worker->priv->display_device != NULL &&
-                   (strncmp (worker->priv->display_device, "/dev/vt/", strlen ("/dev/vt/")) == 0 ||
-                    strcmp  (worker->priv->display_device, "/dev/console") == 0)) {
-                        g_debug ("Logindevperm logout for user %s, device %s",
-                                 worker->priv->username,
-                                 worker->priv->display_device);
-                        (void) di_devperm_logout (worker->priv->display_device);
-                }
-#endif  /* HAVE_LOGINDEVPERM */
-
         } else {
-                void *p;
+                const void *p;
 
                 if ((pam_get_item (worker->priv->pam_handle, PAM_USER, &p)) == PAM_SUCCESS) {
                         gdm_session_auditor_set_username (worker->priv->auditor, (const char *)p);
@@ -1102,7 +1061,7 @@ gdm_session_worker_initialize_pam (GdmSessionWorker *worker,
                 g_set_error (error,
                              GDM_SESSION_WORKER_ERROR,
                              GDM_SESSION_WORKER_ERROR_AUTHENTICATING,
-                             _("error initiating conversation with authentication system - %s"),
+                             _("error initiating conversation with authentication system: %s"),
                              error_code == PAM_ABORT? _("general failure") :
                              error_code == PAM_BUF_ERR? _("out of memory") :
                              error_code == PAM_SYSTEM_ERR? _("application programmer error") :
@@ -1119,7 +1078,7 @@ gdm_session_worker_initialize_pam (GdmSessionWorker *worker,
                         g_set_error (error,
                                      GDM_SESSION_WORKER_ERROR,
                                      GDM_SESSION_WORKER_ERROR_AUTHENTICATING,
-                                     _("error informing authentication system of preferred username prompt - %s"),
+                                     _("error informing authentication system of preferred username prompt: %s"),
                                      pam_strerror (worker->priv->pam_handle, error_code));
                         goto out;
                 }
@@ -1133,7 +1092,7 @@ gdm_session_worker_initialize_pam (GdmSessionWorker *worker,
                         g_set_error (error,
                                      GDM_SESSION_WORKER_ERROR,
                                      GDM_SESSION_WORKER_ERROR_AUTHENTICATING,
-                                     _("error informing authentication system of user's hostname - %s"),
+                                     _("error informing authentication system of user's hostname: %s"),
                                      pam_strerror (worker->priv->pam_handle, error_code));
                         goto out;
                 }
@@ -1148,7 +1107,7 @@ gdm_session_worker_initialize_pam (GdmSessionWorker *worker,
                 g_set_error (error,
                              GDM_SESSION_WORKER_ERROR,
                              GDM_SESSION_WORKER_ERROR_AUTHENTICATING,
-                             _("error informing authentication system of user's console - %s"),
+                             _("error informing authentication system of user's console: %s"),
                              pam_strerror (worker->priv->pam_handle, error_code));
                 goto out;
         }
@@ -1161,7 +1120,7 @@ gdm_session_worker_initialize_pam (GdmSessionWorker *worker,
                 g_set_error (error,
                              GDM_SESSION_WORKER_ERROR,
                              GDM_SESSION_WORKER_ERROR_AUTHENTICATING,
-                             _("error informing authentication system of display string - %s"),
+                             _("error informing authentication system of display string: %s"),
                              pam_strerror (worker->priv->pam_handle, error_code));
                 goto out;
         }
@@ -1176,7 +1135,7 @@ gdm_session_worker_initialize_pam (GdmSessionWorker *worker,
                 g_set_error (error,
                              GDM_SESSION_WORKER_ERROR,
                              GDM_SESSION_WORKER_ERROR_AUTHENTICATING,
-                             _("error informing authentication system of display xauth credentials - %s"),
+                             _("error informing authentication system of display xauth credentials: %s"),
                              pam_strerror (worker->priv->pam_handle, error_code));
                 goto out;
         }
@@ -1216,6 +1175,14 @@ gdm_session_worker_authenticate_user (GdmSessionWorker *worker,
         if (error_code != PAM_SUCCESS) {
                 g_debug ("GdmSessionWorker: authentication returned %d: %s", error_code, pam_strerror (worker->priv->pam_handle, error_code));
 
+                /*
+                 * Do not display a different message for user unknown versus
+                 * a failed password for a valid user.
+                 */
+                if (error_code == PAM_USER_UNKNOWN) {
+                        error_code = PAM_AUTH_ERR;
+                }
+
                 g_set_error (error,
                              GDM_SESSION_WORKER_ERROR,
                              GDM_SESSION_WORKER_ERROR_AUTHENTICATING,
@@ -1294,13 +1261,32 @@ gdm_session_worker_set_environment_variable (GdmSessionWorker *worker,
                                              const char       *key,
                                              const char       *value)
 {
-        /* FIXME: maybe we should use use pam_putenv instead of our
-         * own hash table, so pam can override our choices if it knows
-         * better?
-         */
-        g_hash_table_replace (worker->priv->environment,
-                              g_strdup (key),
-                              g_strdup (value));
+        int error_code;
+        char *environment_entry;
+
+        if (value != NULL) {
+                environment_entry = g_strdup_printf ("%s=%s", key, value);
+        } else {
+                /* empty value means "remove from environment" */
+                environment_entry = g_strdup (key);
+        }
+
+        error_code = pam_putenv (worker->priv->pam_handle,
+                                 environment_entry);
+
+        if (error_code != PAM_SUCCESS) {
+                g_warning ("cannot put %s in pam environment: %s\n",
+                           environment_entry,
+                           pam_strerror (worker->priv->pam_handle, error_code));
+        }
+        g_free (environment_entry);
+}
+
+static char *
+gdm_session_worker_get_environment_variable (GdmSessionWorker *worker,
+                                             const char       *key)
+{
+        return g_strdup (pam_getenv (worker->priv->pam_handle, key));
 }
 
 static void
@@ -1319,9 +1305,9 @@ gdm_session_worker_update_environment_from_passwd_info (GdmSessionWorker *worker
 
 static gboolean
 gdm_session_worker_environment_variable_is_set (GdmSessionWorker *worker,
-                                                const char       *name)
+                                                const char       *key)
 {
-        return g_hash_table_lookup (worker->priv->environment, name) != NULL;
+        return pam_getenv (worker->priv->pam_handle, key) != NULL;
 }
 
 static gboolean
@@ -1345,6 +1331,7 @@ _change_user (GdmSessionWorker  *worker,
         }
 #endif
         worker->priv->uid = uid;
+        worker->priv->gid = gid;
 
         if (setgid (gid) < 0) {
                 return FALSE;
@@ -1390,6 +1377,7 @@ _lookup_passwd_info (const char *username,
          * passwd_entry doesn't potentially get stomped on
          * by a PAM module
          */
+ again:
         passwd_entry = NULL;
 #ifdef HAVE_POSIX_GETPWNAM_R
         errno = getpwnam_r (username,
@@ -1404,8 +1392,10 @@ _lookup_passwd_info (const char *username,
                                    (size_t) aux_buffer_size);
         errno = 0;
 #endif /* !HAVE_POSIX_GETPWNAM_R */
-
-        if (errno != 0) {
+        if (errno == EINTR) {
+                g_debug ("%s", g_strerror (errno));
+                goto again;
+        } else if (errno != 0) {
                 g_warning ("%s", g_strerror (errno));
                 goto out;
         }
@@ -1450,6 +1440,9 @@ gdm_session_worker_accredit_user (GdmSessionWorker  *worker,
 
         ret = FALSE;
 
+        home = NULL;
+        shell = NULL;
+
         if (worker->priv->username == NULL) {
                 g_debug ("GdmSessionWorker: Username not set");
                 error_code = PAM_USER_UNKNOWN;
@@ -1460,8 +1453,6 @@ gdm_session_worker_accredit_user (GdmSessionWorker  *worker,
                 goto out;
         }
 
-        home = NULL;
-        shell = NULL;
         uid = 0;
         gid = 0;
         res = _lookup_passwd_info (worker->priv->username,
@@ -1471,6 +1462,7 @@ gdm_session_worker_accredit_user (GdmSessionWorker  *worker,
                                    &shell);
         if (! res) {
                 g_debug ("GdmSessionWorker: Unable to lookup account info");
+                error_code = PAM_AUTHINFO_UNAVAIL;
                 g_set_error (error,
                              GDM_SESSION_WORKER_ERROR,
                              GDM_SESSION_WORKER_ERROR_GIVING_CREDENTIALS,
@@ -1519,6 +1511,8 @@ gdm_session_worker_accredit_user (GdmSessionWorker  *worker,
         ret = TRUE;
 
  out:
+        g_free (home);
+        g_free (shell);
         if (ret) {
                 g_debug ("GdmSessionWorker: state ACCREDITED");
                 ret = TRUE;
@@ -1531,58 +1525,10 @@ gdm_session_worker_accredit_user (GdmSessionWorker  *worker,
         return ret;
 }
 
-static void
-gdm_session_worker_update_environment_from_pam (GdmSessionWorker *worker)
-{
-        char **environment;
-        gsize i;
-
-        environment = pam_getenvlist (worker->priv->pam_handle);
-
-        for (i = 0; environment[i] != NULL; i++) {
-                char **key_and_value;
-
-                key_and_value = g_strsplit (environment[i], "=", 2);
-
-                gdm_session_worker_set_environment_variable (worker, key_and_value[0], key_and_value[1]);
-
-                g_strfreev (key_and_value);
-        }
-
-        for (i = 0; environment[i]; i++) {
-                free (environment[i]);
-        }
-
-        free (environment);
-}
-
-static void
-gdm_session_worker_fill_environment_array (const char *key,
-                                           const char *value,
-                                           GPtrArray  *environment)
-{
-        char *variable;
-
-        if (value == NULL)
-                return;
-
-        variable = g_strdup_printf ("%s=%s", key, value);
-
-        g_ptr_array_add (environment, variable);
-}
-
 static char **
 gdm_session_worker_get_environment (GdmSessionWorker *worker)
 {
-        GPtrArray *environment;
-
-        environment = g_ptr_array_new ();
-        g_hash_table_foreach (worker->priv->environment,
-                              (GHFunc) gdm_session_worker_fill_environment_array,
-                              environment);
-        g_ptr_array_add (environment, NULL);
-
-        return (char **) g_ptr_array_free (environment, FALSE);
+        return pam_getenvlist (worker->priv->pam_handle);
 }
 
 static void
@@ -1726,25 +1672,6 @@ out:
         return fd;
 }
 
-static void
-_save_user_settings (GdmSessionWorker *worker,
-                     const char       *home_dir)
-{
-        GError *error;
-
-        if (!gdm_session_settings_is_loaded (worker->priv->user_settings)) {
-                return;
-        }
-
-        error = NULL;
-        if (!gdm_session_settings_save (worker->priv->user_settings,
-                                        home_dir, &error)) {
-                g_warning ("could not save session and language settings: %s",
-                           error->message);
-                g_error_free (error);
-        }
-}
-
 static gboolean
 gdm_session_worker_start_user_session (GdmSessionWorker  *worker,
                                        GError           **error)
@@ -1753,31 +1680,9 @@ gdm_session_worker_start_user_session (GdmSessionWorker  *worker,
         pid_t session_pid;
         int   error_code;
 
-        g_debug ("GdmSessionWorker: querying pam for user environment");
-        gdm_session_worker_update_environment_from_pam (worker);
-
         register_ck_session (worker);
 
-        passwd_entry = getpwnam (worker->priv->username);
-
-#ifdef  HAVE_LOGINDEVPERM
-        /*
-         * Only do logindevperm processing if /dev/console or
-         * a device associated with a VT
-         */
-        if (worker->priv->display_device != NULL &&
-           (strncmp (worker->priv->display_device, "/dev/vt/", strlen ("/dev/vt/")) == 0 ||
-            strcmp  (worker->priv->display_device, "/dev/console") == 0)) {
-                g_debug ("Logindevperm login for user %s, device %s",
-                         worker->priv->username,
-                         worker->priv->display_device);
-                (void) di_devperm_login (worker->priv->display_device,
-                                         passwd_entry->pw_uid,
-                                         passwd_entry->pw_gid,
-                                         NULL);
-        }
-#endif  /* HAVE_LOGINDEVPERM */
-
+        gdm_get_pwent_for_name (worker->priv->username, &passwd_entry);
         g_debug ("GdmSessionWorker: opening user session with program '%s'",
                  worker->priv->arguments[0]);
 
@@ -1800,12 +1705,12 @@ gdm_session_worker_start_user_session (GdmSessionWorker  *worker,
                 int    fd;
 
                 if (setuid (worker->priv->uid) < 0) {
-                        g_debug ("GdmSessionWorker: could not reset uid - %s", g_strerror (errno));
+                        g_debug ("GdmSessionWorker: could not reset uid: %s", g_strerror (errno));
                         _exit (1);
                 }
 
                 if (setsid () < 0) {
-                        g_debug ("GdmSessionWorker: could not set pid '%u' as leader of new session and process group - %s",
+                        g_debug ("GdmSessionWorker: could not set pid '%u' as leader of new session and process group: %s",
                                  (guint) getpid (), g_strerror (errno));
                         _exit (2);
                 }
@@ -1814,12 +1719,11 @@ gdm_session_worker_start_user_session (GdmSessionWorker  *worker,
 
                 g_assert (geteuid () == getuid ());
 
-                home_dir = g_hash_table_lookup (worker->priv->environment,
-                                                "HOME");
-
+                home_dir = gdm_session_worker_get_environment_variable (worker, "HOME");
                 if ((home_dir == NULL) || g_chdir (home_dir) < 0) {
                         g_chdir ("/");
                 }
+                g_free (home_dir);
 
                 fd = open ("/dev/null", O_RDWR);
                 dup2 (fd, STDIN_FILENO);
@@ -1830,17 +1734,26 @@ gdm_session_worker_start_user_session (GdmSessionWorker  *worker,
                 dup2 (fd, STDERR_FILENO);
                 close (fd);
 
-                _save_user_settings (worker, home_dir);
+                if (!gdm_session_settings_save (worker->priv->user_settings,
+                                                worker->priv->username)) {
+                        g_warning ("could not save session and language settings");
+                }
+
+                /*
+                 * Reset SIGPIPE to default so that any process in the user
+                 * session get the default SIGPIPE behavior instead of ignoring
+                 * SIGPIPE.
+                 */
+                signal (SIGPIPE, SIG_DFL);
 
                 gdm_session_execute (worker->priv->arguments[0],
                                      worker->priv->arguments,
                                      environment,
                                      TRUE);
 
-                g_debug ("GdmSessionWorker: child '%s' could not be started - %s",
+                g_debug ("GdmSessionWorker: child '%s' could not be started: %s",
                          worker->priv->arguments[0],
                          g_strerror (errno));
-                g_strfreev (environment);
 
                 _exit (127);
         }
@@ -1870,7 +1783,7 @@ gdm_session_worker_open_user_session (GdmSessionWorker  *worker,
 {
         int error_code;
 
-        g_assert (worker->priv->state == GDM_SESSION_WORKER_STATE_ACCREDITED);
+        g_assert (worker->priv->state == GDM_SESSION_WORKER_STATE_ACCOUNT_DETAILS_FETCHED);
         g_assert (geteuid () == 0);
 
         error_code = pam_open_session (worker->priv->pam_handle, 0);
@@ -2007,14 +1920,6 @@ gdm_session_worker_set_language_name (GdmSessionWorker *worker,
                                                 language_name);
 }
 
-static void
-gdm_session_worker_set_layout_name (GdmSessionWorker *worker,
-                                    const char       *layout_name)
-{
-        gdm_session_settings_set_layout_name (worker->priv->user_settings,
-                                              layout_name);
-}
-
 static void
 on_set_language_name (GdmSessionWorker *worker,
                       DBusMessage      *message)
@@ -2037,28 +1942,6 @@ on_set_language_name (GdmSessionWorker *worker,
         }
 }
 
-static void
-on_set_layout_name (GdmSessionWorker *worker,
-                    DBusMessage      *message)
-{
-        DBusError   error;
-        const char *layout_name;
-        dbus_bool_t res;
-
-        dbus_error_init (&error);
-        res = dbus_message_get_args (message,
-                                     &error,
-                                     DBUS_TYPE_STRING, &layout_name,
-                                     DBUS_TYPE_INVALID);
-        if (res) {
-                g_debug ("GdmSessionWorker: layout name set to %s", layout_name);
-                gdm_session_worker_set_layout_name (worker, layout_name);
-        } else {
-                g_warning ("Unable to get arguments: %s", error.message);
-                dbus_error_free (&error);
-        }
-}
-
 static void
 on_saved_language_name_read (GdmSessionWorker *worker)
 {
@@ -2071,18 +1954,6 @@ on_saved_language_name_read (GdmSessionWorker *worker)
         g_free (language_name);
 }
 
-static void
-on_saved_layout_name_read (GdmSessionWorker *worker)
-{
-        char *layout_name;
-
-        layout_name = gdm_session_settings_get_layout_name (worker->priv->user_settings);
-        send_dbus_string_method (worker->priv->connection,
-                                 "SavedLayoutNameRead",
-                                 layout_name);
-        g_free (layout_name);
-}
-
 static void
 on_saved_session_name_read (GdmSessionWorker *worker)
 {
@@ -2108,11 +1979,6 @@ do_setup (GdmSessionWorker *worker)
                                   G_CALLBACK (on_saved_language_name_read),
                                   worker);
 
-        g_signal_connect_swapped (worker->priv->user_settings,
-                                  "notify::layout-name",
-                                  G_CALLBACK (on_saved_layout_name_read),
-                                  worker);
-
         g_signal_connect_swapped (worker->priv->user_settings,
                                   "notify::session-name",
                                   G_CALLBACK (on_saved_session_name_read),
@@ -2228,6 +2094,43 @@ do_accredit (GdmSessionWorker *worker)
         send_dbus_void_method (worker->priv->connection, "Accredited");
 }
 
+static void
+on_settings_is_loaded_changed (GdmSessionWorker *worker)
+{
+        if (!gdm_session_settings_is_loaded (worker->priv->user_settings)) {
+                return;
+        }
+
+        if (worker->priv->state == GDM_SESSION_WORKER_STATE_ACCREDITED) {
+                worker->priv->state = GDM_SESSION_WORKER_STATE_ACCOUNT_DETAILS_FETCHED;
+                queue_state_change (worker);
+        }
+
+        g_signal_handlers_disconnect_by_func (G_OBJECT (worker->priv->user_settings),
+                                              G_CALLBACK (on_settings_is_loaded_changed),
+                                              worker);
+}
+
+static void
+do_fetch_account_details (GdmSessionWorker *worker)
+{
+        g_assert (worker->priv->state == GDM_SESSION_WORKER_STATE_ACCREDITED);
+
+        if (gdm_session_settings_is_loaded (worker->priv->user_settings)) {
+                worker->priv->state = GDM_SESSION_WORKER_STATE_ACCOUNT_DETAILS_FETCHED;
+                queue_state_change (worker);
+                return;
+        }
+
+        g_signal_connect (G_OBJECT (worker->priv->user_settings),
+                          "notify::is-loaded",
+                          G_CALLBACK (on_settings_is_loaded_changed),
+                          worker);
+
+        gdm_session_settings_load (worker->priv->user_settings,
+                                   worker->priv->username);
+}
+
 static void
 do_open_session (GdmSessionWorker *worker)
 {
@@ -2238,13 +2141,13 @@ do_open_session (GdmSessionWorker *worker)
         res = gdm_session_worker_open_user_session (worker, &error);
         if (! res) {
                 send_dbus_string_method (worker->priv->connection,
-                                         "StartFailed",
+                                         "OpenFailed",
                                          error->message);
                 g_error_free (error);
                 return;
         }
 
-        queue_state_change (worker);
+        send_dbus_void_method (worker->priv->connection, "SessionOpened");
 }
 
 static void
@@ -2291,6 +2194,9 @@ get_state_name (int state)
         case GDM_SESSION_WORKER_STATE_ACCREDITED:
                 name = "ACCREDITED";
                 break;
+        case GDM_SESSION_WORKER_STATE_ACCOUNT_DETAILS_FETCHED:
+                name = "ACCOUNT_DETAILS_FETCHED";
+                break;
         case GDM_SESSION_WORKER_STATE_SESSION_OPENED:
                 name = "SESSION_OPENED";
                 break;
@@ -2329,6 +2235,9 @@ state_change_idle (GdmSessionWorker *worker)
         case GDM_SESSION_WORKER_STATE_ACCREDITED:
                 do_accredit (worker);
                 break;
+        case GDM_SESSION_WORKER_STATE_ACCOUNT_DETAILS_FETCHED:
+                do_fetch_account_details (worker);
+                break;
         case GDM_SESSION_WORKER_STATE_SESSION_OPENED:
                 do_open_session (worker);
                 break;
@@ -2360,7 +2269,7 @@ on_start_program (GdmSessionWorker *worker,
         const char *text;
         dbus_bool_t res;
 
-        if (worker->priv->state != GDM_SESSION_WORKER_STATE_ACCREDITED) {
+        if (worker->priv->state != GDM_SESSION_WORKER_STATE_SESSION_OPENED) {
                 g_debug ("GdmSessionWorker: ignoring spurious start program while in state %s", get_state_name (worker->priv->state));
                 return;
         }
@@ -2518,6 +2427,18 @@ on_establish_credentials (GdmSessionWorker *worker,
         queue_state_change (worker);
 }
 
+static void
+on_open_session (GdmSessionWorker *worker,
+                 DBusMessage      *message)
+{
+        if (worker->priv->state != GDM_SESSION_WORKER_STATE_ACCREDITED) {
+                g_debug ("GdmSessionWorker: ignoring spurious open session for user while in state %s", get_state_name (worker->priv->state));
+                return;
+        }
+
+        queue_state_change (worker);
+}
+
 static void
 on_reauthenticate (GdmSessionWorker *worker,
                    DBusMessage      *message)
@@ -2590,6 +2511,8 @@ worker_dbus_handle_message (DBusConnection *connection,
                 on_authorize (worker, message);
         } else if (dbus_message_is_signal (message, GDM_SESSION_DBUS_INTERFACE, "EstablishCredentials")) {
                 on_establish_credentials (worker, message);
+        } else if (dbus_message_is_signal (message, GDM_SESSION_DBUS_INTERFACE, "OpenSession")) {
+                on_open_session (worker, message);
         } else if (dbus_message_is_signal (message, GDM_SESSION_DBUS_INTERFACE, "StartProgram")) {
                 on_start_program (worker, message);
         } else if (dbus_message_is_signal (message, GDM_SESSION_DBUS_INTERFACE, "Reauthenticate")) {
@@ -2602,8 +2525,6 @@ worker_dbus_handle_message (DBusConnection *connection,
                 on_set_environment_variable (worker, message);
         } else if (dbus_message_is_signal (message, GDM_SESSION_DBUS_INTERFACE, "SetLanguageName")) {
                 on_set_language_name (worker, message);
-        } else if (dbus_message_is_signal (message, GDM_SESSION_DBUS_INTERFACE, "SetLayoutName")) {
-                on_set_layout_name (worker, message);
         } else if (dbus_message_is_signal (message, GDM_SESSION_DBUS_INTERFACE, "SetSessionName")) {
                 on_set_session_name (worker, message);
         } else {
@@ -2710,10 +2631,6 @@ gdm_session_worker_init (GdmSessionWorker *worker)
 {
 
         worker->priv = GDM_SESSION_WORKER_GET_PRIVATE (worker);
-        worker->priv->environment = g_hash_table_new_full (g_str_hash,
-                                                           g_str_equal,
-                                                           (GDestroyNotify) g_free,
-                                                           (GDestroyNotify) g_free);
 }
 
 static void
@@ -2741,20 +2658,14 @@ gdm_session_worker_finalize (GObject *object)
 
         gdm_session_worker_unwatch_child (worker);
 
-        if (worker->priv->username != NULL) {
-                g_free (worker->priv->username);
-                worker->priv->username = NULL;
-        }
-
-        if (worker->priv->arguments != NULL) {
-                g_strfreev (worker->priv->arguments);
-                worker->priv->arguments = NULL;
-        }
-
-        if (worker->priv->environment != NULL) {
-                g_hash_table_destroy (worker->priv->environment);
-                worker->priv->environment = NULL;
-        }
+        g_free (worker->priv->service);
+        g_free (worker->priv->x11_display_name);
+        g_free (worker->priv->x11_authority_file);
+        g_free (worker->priv->display_device);
+        g_free (worker->priv->hostname);
+        g_free (worker->priv->username);
+        g_free (worker->priv->server_address);
+        g_strfreev (worker->priv->arguments);
 
         G_OBJECT_CLASS (gdm_session_worker_parent_class)->finalize (object);
 }
